# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image statistics & histograms."""

import numpy as np

from . import params
from . import helpers

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  def histograms(self, channels = "RGBL", nbins = None, recompute = False):
    """Compute histograms of selected channels of the image.

    The histograms are both returned and embedded in the object as self.hists. Histograms already registered in self.hists
    are not recomputed unless required.

    Args:
      channels (str, optional): A combination of the keys "R" (for red), "G" (for green), "B" (for blue),  "V" (for HSV value),
        "S" (for HSV saturation), and "L" (for luma). For a HSV image, only the histograms of the value and saturation can be
        computed. If channels ends with a "*", it gets appended with the keys already computed and stored in self.hists.
        Default is "RGBL".
      nbins (int, optional): Number of bins within [0, 1] in the histograms. Set to `equimage.params.maxhistbins` if negative,
        and computed from the image statistics using Scott's rule if zero. If None, defaults to `equimage.params.defhistbins`.
      recompute (bool, optional): If False (default), the histograms already registered in self.hists are not recomputed
        (provided they match nbins). If True, all histograms are recomputed.

    Returns:
      dict: hists[key] for key in channels, with:

        - hists[key].name = channel name ("Red", "Green", "Blue", "Value", "Luma" or "Saturation", provided for convenience).
        - hists[key].nbins = number of bins within [0, 1].
        - hists[key].edges = histogram bins edges.
        - hists[key].counts = histogram bins counts.
        - hists[key].color = suggested line color for plots.
    """
    if nbins is None: nbins = params.defhistbins
    if nbins == 0:
      if not recompute and hasattr(self, "hists"): # Retrieve the number of bins from the existing histograms.
        nbins = list(self.hists.values())[0].nbins
      else: # Compute the number of bins using Scott's rule.
        width, height = self.get_size()
        npixels = width*height
        if self.colormodel == "RGB":
          channel = (self.image[0]+self.image[1]+self.image[2])/3.
        elif self.colormodel == "HSV":
          channel = self.image[2]
        elif self.colormodel == "gray":
          channel = self.image[0]
        else:
          self.color_model_error()
        nbins = int(np.ceil(npixels**(1./3.)/(3.5*np.std(channel))))
    elif nbins < 0:
      nbins = params.maxhistbins
    nbins = min(max(nbins, 16), params.maxhistbins)
    if not hasattr(self, "hists"): self.hists = {} # Register empty histograms in the object, if none already computed.
    if len(channels) > 0: # Append missing keys if channels ends with a "*".
      if channels[-1] == "*":
        channels = channels[:-1]
        for key in self.hists.keys():
          if not key in channels:
            channels += key
    hists = {}
    for key in channels:
      if key in hists: # Already selected.
        print(f"Warning, channel '{key}' selected twice or more...")
        continue
      if not recompute and key in self.hists: # Already computed.
        if self.hists[key].nbins == nbins:
          hists[key] = self.hists[key]
          continue
      if key == "R":
        self.check_color_model("RGB", "gray")
        name = "Red"
        color = "red"
        channel = self.image[0]
      elif key == "G":
        self.check_color_model("RGB", "gray")
        name = "Green"
        color = "green"
        channel = self.image[1] if self.colormodel == "RGB" else self.image[0]
      elif key == "B":
        self.check_color_model("RGB", "gray")
        name = "Blue"
        color = "blue"
        channel = self.image[2] if self.colormodel == "RGB" else self.image[0]
      elif key == "V":
        name = "Value"
        color = "darkslategray"
        channel = self.value()
      elif key == "S":
        name = "Saturation"
        color = "orange"
        channel = self.saturation()
      elif key == "L":
        name = "Luma"
        color = "lightslategray"
        channel = self.luma()
      else:
        raise ValueError(f"Error, unknown channel '{key}'.")
      mmin = np.floor(np.min(channel)*nbins)
      mmax = np.ceil (np.max(channel)*nbins)
      mbins = int(round(mmax-mmin))
      hists[key] = helpers.Container()
      hists[key].name = name
      hists[key].nbins = nbins
      hists[key].counts, hists[key].edges = np.histogram(channel, bins = mbins, range = (mmin/nbins, mmax/nbins), density = False)
      hists[key].color = color
    self.hists = hists
    return hists

  def statistics(self, channels = "RGBL", exclude01 = None, recompute = False):
    """Compute statistics of selected channels of the image.

    The statistics are both returned and embedded in the object as self.stats. Statistics already registered in self.stats are
    not recomputed unless required.

    Args:
      channels (str, optional): A combination of the keys "R" (for red), "G" (for green), "B" (for blue), "V" (for HSV value),
        "S" (for HSV saturation), and "L" (for luma). For a HSV image, only the statistics of the value and saturation can be
        computed. If channels ends with a "*", it gets appended with the keys already computed and stored in self.stats.
        Default is "RGBL".
      exclude01 (bool, optional): If True, exclude pixels <= 0 or >= 1 from the median and percentiles.
        Defaults to `equimage.params.exclude01` if None.
      recompute (bool, optional): If False (default), the statistics already registered in self.stats are not recomputed.
        If True, all statistics are recomputed.

    Returns:
      dict: stats[key] for key in channels, with:

        - stats[key].name = channel name ("Red", "Green", "Blue", "Value", "Luma" or "Saturation", provided for convenience).
        - stats[key].width = image width (provided for convenience).
        - stats[key].height = image height (provided for convenience).
        - stats[key].npixels = number of image pixels = image width*image height (provided for convenience).
        - stats[key].minimum = minimum level.
        - stats[key].maximum = maximum level.
        - stats[key].percentiles = (pr25, pr50, pr75) = the 25th, 50th and 75th percentiles.
        - stats[key].median = pr50 = median level.
        - stats[key].zerocount = number of pixels <= 0.
        - stats[key].outcount = number of pixels > 1 (out-of-range).
        - stats[key].exclude01 = True if pixels >= 0 or <= 1 have been excluded from the median and percentiles, False otherwise.
        - stats[key].color = suggested text color for display.
    """
    if exclude01 is None: exclude01 = params.exclude01
    if not hasattr(self, "stats"): self.stats = {} # Register empty statistics in the object, if none already computed.
    if len(channels) > 0: # Append missing keys if channels ends with a "*".
      if channels[-1] == "*":
        channels = channels[:-1]
        for key in self.stats.keys():
          if not key in channels:
            channels += key
    width, height = self.get_size()
    npixels = width*height
    stats = {}
    for key in channels:
      if key in stats: # Already selected.
        print(f"Warning, channel '{key}' selected twice or more...")
        continue
      if not recompute and key in self.stats: # Already computed.
        if self.stats[key].exclude01 == exclude01:
          stats[key] = self.stats[key]
          continue
      if key == "R":
        self.check_color_model("RGB", "gray")
        name = "Red"
        color = "red"
        channel = self.image[0]
      elif key == "G":
        self.check_color_model("RGB", "gray")
        name = "Green"
        color = "green"
        channel = self.image[1] if self.colormodel == "RGB" else self.image[0]
      elif key == "B":
        self.check_color_model("RGB", "gray")
        name = "Blue"
        color = "blue"
        channel = self.image[2] if self.colormodel == "RGB" else self.image[0]
      elif key == "V":
        name = "Value"
        color = "darkslategray"
        channel = self.value()
      elif key == "S":
        name = "Saturation"
        color = "orange"
        channel = self.saturation()
      elif key == "L":
        name = "Luma"
        color = "lightslategray"
        channel = self.luma()
      else:
        raise ValueError(f"Error, unknown channel '{key}'.")
      stats[key] = helpers.Container()
      stats[key].name = name
      stats[key].width = width
      stats[key].height = height
      stats[key].npixels = npixels
      stats[key].minimum = np.min(channel)
      stats[key].maximum = np.max(channel)
      if exclude01:
        mask = (channel >= params.IMGTOL) & (channel <= 1.-params.IMGTOL)
        stats[key].percentiles = np.percentile(channel[mask], [25., 50., 75.]) if np.any(mask) else None
      else:
        stats[key].percentiles = np.percentile(channel, [25., 50., 75.])
      stats[key].median = stats[key].percentiles[1] if stats[key].percentiles is not None else None
      stats[key].zerocount = np.sum(channel < params.IMGTOL)
      stats[key].outcount = np.sum(channel > 1.+params.IMGTOL)
      stats[key].exclude01 = exclude01
      stats[key].color = color
    self.stats = stats
    return stats



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Jupyter-lab interface parameters."""

import numpy as np

# Images downsampling rate for display.
# Only images[::sampling, ::sampling] are shown to speed up operations.

sampling = 1

# Figure size and margins (pixels).

maxwidth = 1024 # Maximum width of the displayed image.

lmargin = 96  # Left margin.
rmargin = 160 # Right margin.
bmargin = 48  # Bottom margin.
tmargin = 48  # Top margin.

rowheight = 25 # Height of table rows.

# Color of marker lines.

mlinecolor = "mediumslateblue"

# Colors for shadows, highlights and differences.

shadowcolor = np.array([1., .5, 0.])
highlightcolor = np.array([1., 1., 0.])
diffcolor = np.array([1., 1., 0.])

# Setters.

def set_image_sampling(s):
  """Set image downsampling rate for display.

  Args:
    s (int): The image downsampling rate in pixels.
  """
  global sampling
  sampling = s

def set_figure_max_width(w):
  """Set maximum figure width.

  Args:
    w (int): The maximum figure width in pixels.
  """
  global maxwidth
  maxwidth = w

def set_figure_margins(left = None, right = None, bottom = None, top = None):
  """Set or update figure margins.

  Args:
    left (int, optional): The left margin in pixels [if not None (default)].
    right (int, optional): The right margin in pixels [if not None (default)].
    bottom (int, optional): The bottom margin in pixels [if not None (default)].
    top (int, optional): The top margin in pixels [if not None (default)].
  """
  global lmargin, rmargin, tmargin, bmargin
  if left is not None: lmargin = left
  if right is not None: rmargin = right
  if bottom is not None: bmargin = bottom
  if top is not None: tmargin = top

def set_table_row_height(h):
  """Set table row height.

  Args:
    h (int): The table row height in pixels.
  """
  global rowheight
  rowheight = h



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image class."""

import copy
import numpy as np

from . import params
from . import image_colorspaces
from . import image_utils
from . import image_geometry
from . import image_colors
from . import image_stretch
from . import image_filters
from . import image_skimage
from . import image_masks
from . import image_stats
from . import image_editors
from . import image_io

class Image(np.lib.mixins.NDArrayOperatorsMixin,
            image_colorspaces.Mixin, image_utils.Mixin, image_geometry.Mixin,
            image_colors.Mixin, image_stretch.Mixin, image_filters.Mixin, image_skimage.Mixin,
            image_masks.Mixin, image_stats.Mixin, image_editors.Mixin, image_io.Mixin):
  """Image class.

  The image is stored as self.image, a numpy.ndarray with dtype params.IMGTYPE = np.float32 or np.float64.
  Color images are stored as arrays with shape (3, height, width) and grayscale images as arrays with
  shape (1, height, width). The leading axis spans the color channels, and the last two the height
  and width of the image.

  The class embeds colorspace and colormodel attributes for the color space and model of the image.

  The colorspace attribute can be:

    - "lRGB" for the linear RGB color space.
    - "sRGB" for the sRGB color space.

  The colormodel attribute can be:

    - "RGB": the 3 channels of the image are the red, blue, and green values within [0, 1].
    - "HSV": the 3 channels of the image are the hue, value, and saturation within [0, 1].
    - "gray": grayscale image with one single channel within [0, 1].

  The default color space is sRGB and the default color model is RGB.
  """

  ################
  # Constructor. #
  ################

  def __init__(self, image, colorspace = "sRGB", colormodel = "RGB", channels = 0):
    """Initialize a new Image object with the input image.

    Args:
      image: The input image (numpy.ndarray or Image).
      colorspace (str, optional): The image color space (default "sRGB").
        Can be "lRGB" (linear RGB color space) or "sRGB" (sRGB color space).
      colormodel (str, optional): The image color model (default "RGB").
        Can be "RGB" (RGB image), "HSV" (HSV image) or "gray (grayscale image).
      channels (int, optional): The position of the channel axis for color images (default 0).
    """
    # Check color space and model.
    if colorspace not in ["lRGB", "sRGB"]:
      raise ValueError(f"Error, the color space must either be 'lRGB' or 'sRGB' (got {colorspace}).")
    if colormodel not in ["RGB", "HSV", "gray"]:
      raise ValueError(f"Error, the color model must either be 'RGB', 'HSV' or 'gray' (got {colormodel}).")
    # Convert the input image into an array.
    image = np.asarray(image, dtype = params.IMGTYPE)
    # Validate the image.
    if image.ndim == 2:
      colormodel = "gray"  # Enforce colormodel = "gray".
      image = np.expand_dims(image, axis = 0)
    elif image.ndim == 3:
      if channels != 0: image = np.moveaxis(image, channels, 0)
      nc = image.shape[0]
      if nc == 1:
        colormodel = "gray" # Enforce colormodel = "gray".
      elif nc == 3:
        if colormodel == "gray":
          raise ValueError(f"Error, a grayscale image must have one single channel (found {nc}).")
      else:
        raise ValueError(f"Error, an image must have 1 or 3 channels (found {nc}).")
    else:
      raise ValueError(f"Error, an image must have 2 or 3 dimensions (found {image.ndim}).")
    # Register image, color space and model.
    self.image = image
    self.colorspace = colorspace
    self.colormodel = colormodel

  def newImage(self, image, **kwargs):
    """Return a new Image object with the input image.

    Args:
      image (numpy.ndarray): The input image.
      colorspace (str, optional): The image color space (default self.colorspace).
        Can be "lRGB" (linear RGB color space) or "sRGB" (sRGB color space).
      colormodel (str, optional): The image color model (default self.colormodel).
        Can be "RGB" (RGB image), "HSV" (HSV image) or "gray (grayscale image).

    Returns:
      Image: The new Image object.
    """
    colorspace = kwargs.pop("colorspace", self.colorspace)
    colormodel = kwargs.pop("colormodel", self.colormodel)
    if kwargs: print("Discarding extra keyword arguments in Image.newImage...")
    return Image(image, colorspace = colorspace, colormodel = colormodel)

  def copy(self):
    """Return a copy of the object.

    Returns:
      Image: A (deep) copy of the object.
    """
    return copy.deepcopy(self)

  ######################
  # Object management. #
  ######################

  def __repr__(self):
    """Return the object representation."""
    return f"{self.__class__.__name__}(colorspace = {self.colorspace}, colormodel = {self.colormodel}, size = {self.image.shape[2]}x{self.image.shape[1]} pixels)"

  def __array__(self, dtype = None, copy = None):
    """Expose the object as an numpy.ndarray."""
    return np.array(self.image, dtype = dtype, copy = copy)

  def __array_ufunc__(self, ufunc, method, *args, **kwargs):
    """Apply numpy ufuncs to the object."""
    if method != "__call__": return
    inputs = []
    mixed = False
    reference = None
    for arg in args:
      if isinstance(arg, Image):
        if reference is None:
          reference = arg
        else:
          if arg.colorspace != reference.colorspace or arg.colormodel != reference.colormodel and not mixed:
            print("Warning ! This operation mixes images with different color spaces or models !..")
            mixed = True
        inputs.append(arg.image)
      else:
        inputs.append(arg)
    output = ufunc(*inputs, **kwargs)
    if isinstance(output, np.ndarray):
      if output.ndim == 0: # Is output actually a scalar ?
        return output[()]
      else:
        if not mixed and output.shape == reference.image.shape:
          return Image(output, colorspace = reference.colorspace, colormodel = reference.colormodel)
        else:
          return output
    else:
      return output

  def __array_function__(self, func, types, args, kwargs):
    """Apply numpy array functions to the object."""
    inputs = []
    mixed = False
    reference = None
    for arg in args:
      if isinstance(arg, Image):
        if reference is None:
          reference = arg
        else:
          if arg.colorspace != reference.colorspace or arg.colormodel != reference.colormodel and not mixed:
            print("Warning ! This operation mixes images with different color spaces or models !..")
            mixed = True
        inputs.append(arg.image)
      else:
        inputs.append(arg)
    output = func(*inputs, **kwargs)
    if isinstance(output, np.ndarray):
      if output.ndim == 0: # Is output actually a scalar ?
        return output[()]
      else:
        if not mixed and output.shape == reference.image.shape:
          return Image(output, colorspace = reference.colorspace, colormodel = reference.colormodel)
        else:
          return output
    else:
      return output

  ##################
  # Image queries. #
  ##################

  def get_image(self, channels = 0, copy = False):
    """Return the image data.

    Args:
      channels (int, optional): The position of the channel axis (default 0).
      copy (bool, optional): If True, return a copy of the image data;
                             If False (default), return a view.

    Returns:
      numpy.ndarray: The image data.
    """
    image = self.image
    if channels != 0: image = np.moveaxis(image, 0, channels)
    return image.copy() if copy else image

  def get_shape(self):
    """Return the shape of the image data.

    Returns:
      tuple: (number of channels, height of the image in pixels, width of the image in pixels).
    """
    return self.image.shape

  def get_size(self):
    """Return the width and height of the image.

    Returns:
      tuple: (width, height) of the image in pixels.
    """
    return self.image.shape[2], self.image.shape[1]

  def get_nc(self):
    """Return the number of channels of the image.

    Returns:
      int: The number of channels of the image.
    """
    return self.image.shape[0]

  ######################
  # Image conversions. #
  ######################

  def int8(self):
    """Return the image as a (height, width, channels) array of 8 bits integers in the range [0, 255].

    Returns:
      numpy.ndarray: The image as a (height, width, channels) array of 8 bits integers in the range [0, 255].
    """
    image = self.get_image(channels = -1)
    data = np.clip(image*255, 0, 255)
    return np.rint(data).astype("uint8")

  def int16(self):
    """Return the image as a (height, width, channels) array of 16 bits integers in the range [0, 65535].

    Returns:
      numpy.ndarray: The image as a (height, width, channels) array of 16 bits integers in the range [0, 65535].
    """
    image = self.get_image(channels = -1)
    data = np.clip(image*65535, 0, 65535)
    return np.rint(data).astype("uint16")

  def int32(self):
    """Return the image as a (height, width, channels) array of 32 bits integers in the range [0, 4294967295].

    Returns:
      numpy.ndarray: The image as a (height, width, channels) array of 32 bits integers in the range [0, 4294967295].
    """
    image = self.get_image(channels = -1)
    data = np.clip(image*4294967295, 0, 4294967295)
    return np.rint(data).astype("uint32")



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Utils for Jupyter-lab interface."""

import base64
from io import BytesIO
from PIL import Image as PILImage
import numpy as np

from . import params

from ..equimage import Image
from ..equimage import params as eqparams

def get_image_size(image):
  """Return the width and height of the input image.

  Args:
    image: An Image object or a numpy.ndarray with dimensions (height, width, 3) (for
      color images), (height, width, 1) or (height, width) (for grayscale images).

  Returns:
    A tuple (width, height) of the image in pixels.
  """
  if issubclass(type(image), Image):
    return image.get_size()
  elif issubclass(type(image), np.ndarray):
    if image.ndim in [2, 3]: return image.shape[1], image.shape[0]
  raise ValueError(f"Error, {image} is not a valid image.")

def prepare_images(images, sampling = -1, copy = False):
  """Prepare images for plotly and Dash.

  Returns all images as numpy.ndarrays with dimensions (height, width, 3) (for color images),
  or (height, width) (for grayscale images).

  Args:
    images: A single/tuple/list of Image object(s) or numpy.ndarrays with dimensions (height, width, 3)
      (for color images), (height, width, 1) or (height, width) (for grayscale images).
    sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
      Only images[::sampling, ::sampling] are processed, to speed up operations.
    copy (bool, optional): If False (default), the output images are (when possible) views of the
      original images; If True, they are always copies.

  Returns:
    A single/tuple/list of float numpy.ndarrays with dimensions (height/sampling, width/sampling, 3) (for
    color images) or (height/sampling, width/sampling) (for grayscale images).
  """

  def prepare(image, copy):
    """Prepare input image."""
    if issubclass(type(image), Image):
      image = image.get_image(channels = -1)
    elif issubclass(type(image), np.ndarray):
      valid = image.ndim in [2, 3]
      if image.ndim == 3: valid = image.shape[2] in [1, 3]
      valid = valid and image.dtype in [np.float32, np.float64]
      if not valid: raise ValueError(f"Error, image {image} is not a valid image.")
    if image.ndim == 3 and image.shape[2] == 1: image = np.squeeze(image, axis = 2)
    if sampling > 1: image = image[::sampling, ::sampling]
    return image.copy() if copy else image

  if sampling <= 0: sampling = params.sampling
  if not isinstance(images, (tuple, list)): return prepare(images, copy)
  return type(images)(prepare(image, copy) for image in images)

def prepare_images_as_b64strings(images, sampling = -1, compression = 4):
  """Prepare images for plotly and Dash as PNGs encoded in base64 strings.

  Returns all images as PNGs encoded in base64 strings.

  Args:
    images: A single/tuple/list of Image object(s) or numpy.ndarrays with dimensions (height, width, 3)
      (for color images), (height, width, 1) or (height, width) (for grayscale images).
    sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
      Only images[::sampling, ::sampling] are processed, to speed up operations.
    compression (int, optional): PNG compression level (default 4).

  Returns:
    A single/tuple/list of images as PNGs encoded in base64 strings.
  """

  def prepare(image):
    """Prepare input image."""
    if issubclass(type(image), Image):
      image = image.get_image(channels = -1)
    elif issubclass(type(image), np.ndarray):
      valid = image.ndim in [2, 3]
      if image.ndim == 3: valid = image.shape[2] in [1, 3]
      valid = valid and image.dtype in [np.float32, np.float64]
      if not valid: raise ValueError(f"Error, image {image} is not a valid image.")
    if image.ndim == 3 and image.shape[2] == 1: image = np.squeeze(image, axis = 2)
    PILimg = PILImage.fromarray(np.clip(image[::sampling, ::sampling]*255, 0, 255).astype("uint8"))
    buffer = BytesIO()
    PILimg.save(buffer, format = "PNG", compress_level = compression)
    return "data:image/png;base64,"+base64.b64encode(buffer.getvalue()).decode("utf-8")

  if sampling <= 0: sampling = params.sampling
  if not isinstance(images, (tuple, list)): return prepare(images)
  return type(images)(prepare(image) for image in images)

def filter(image, channels):
  """Filter the channels of a RGB image.

  Returns a copy of the image with selected red/green/blue channels set to zero.

  Args:
    image: The image (Image object or numpy.ndarray with dimension (height, width, 3)).
    channels (str): The *displayed* channels. A combination of the letters "R" (red),
      "G" (green), and "B" (blue).

  Returns:
    numpy.ndarray: A copy of the image as an array with dimensions (height, width, 3) and the
    non-displayed channels set to zero.
  """
  selected = np.array([False, False, False])
  for c in channels:
    if c == "R":
      selected[0] = True
    elif c == "G":
      selected[1] = True
    elif c == "B":
      selected[2] = True
    else:
      raise ValueError(f"Error, unknown channel {c}.")
  image = prepare_images(image, sampling = 1, copy = True)
  if image.ndim != 3: raise ValueError("Error, the input must be a RGB (not a grayscale) image.""")
  image[:, :, ~selected] = 0.
  return image

def shadowed(image, reference = None):
  """Highlight black pixels in an image.

  Highlight black pixels on the input image with color `jupyter.params.shadowcolor`.
  If a reference image is provided, highlight pixels black on both input and reference images
  with color 0.5 * `jupyter.params.shadowcolor`.

  Args:
    image: The image (Image object or numpy.ndarray).
    reference (optional): The reference image (Image object or numpy.ndarray, default None).

  Returns:
    numpy.ndarray: A copy of the image as an array with dimensions (height, width, 3) and the
    black pixels highlighted with color `jupyter.params.shadowcolor`.
  """
  image = prepare_images(image, sampling = 1, copy = True)
  if image.ndim == 2: image = np.expand_dims(image, axis = -1)
  imgmask = np.all(image < eqparams.IMGTOL, axis = 2)
  if image.shape[2] == 1: image = np.repeat(image, 3, axis = 2)
  image[imgmask, :] = params.shadowcolor
  if reference is not None:
    reference = prepare_images(reference, sampling = 1)
    if reference.shape[0:2] != image.shape[0:2]:
      print("Warning, image and reference have different sizes !")
      return image
    if reference.ndim == 2: reference = np.expand_dims(reference, -1)
    refmask = np.all(reference < eqparams.IMGTOL, axis = 2)
    image[imgmask & refmask, :] = .5*params.shadowcolor
  return image

def highlighted(image, reference = None):
  """Highlight saturated pixels in an image.

  A pixel is saturated if at least one channel is >= 1.
  Show pixels saturated on the input image with color `jupyter.params.highlightcolor`.
  If a reference image is provided, show pixels saturated on both input and reference images
  with color 0.5 * `jupyter.params.highlightcolor`.

  Args:
    image: The image (Image object or numpy.ndarray).
    reference (optional): The reference image (Image object or numpy.ndarray, default None).

  Returns:
    numpy.ndarray: A copy of the image as an array with dimensions (height, width, 3) and the saturated
    pixels highlighted with color `jupyter.params.highlightcolor`.
  """
  image = prepare_images(image, sampling = 1, copy = True)
  if image.ndim == 2: image = np.expand_dims(image, axis = -1)
  imgmask = np.any(image > 1.-eqparams.IMGTOL, axis = 2)
  if image.shape[2] == 1: image = np.repeat(image, 3, axis = 2)
  image[imgmask, :] = params.highlightcolor
  if reference is not None:
    reference = prepare_images(reference, sampling = 1)
    if reference.shape[0:2] != image.shape[0:2]:
      print("Warning, image and reference have different sizes !")
      return image
    if reference.ndim == 2: reference = np.expand_dims(reference, -1)
    refmask = np.any(reference > 1.-eqparams.IMGTOL, axis = 2)
    image[imgmask & refmask, :] = .5*params.highlightcolor
  return image

def differences(image, reference):
  """Highlight differences between an image and a reference.

  Args:
    image: The image (Image object or numpy.ndarray).
    reference: The reference image (Image object or numpy.ndarray).

  Returns:
    numpy.ndarray: A copy of the image as an array with dimensions (height, width, 3)
    and the differences with the reference highlighted with color `jupyter.params.diffcolor`.
  """
  image = prepare_images(image, sampling = 1, copy = True)
  reference = prepare_images(reference, sampling = 1)
  if image.shape != reference.shape:
    raise ValueError("Error, image and reference have different sizes/number of channels !")
  if image.ndim == 2: image = np.expand_dims(image, axis = -1)
  if reference.ndim == 2: reference = np.expand_dims(reference, -1)
  mask = np.any(image != reference, axis = 2)
  if image.shape[2] == 1: image = np.repeat(image, 3, axis = 2)
  image[mask, :] = params.diffcolor
  return image



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""External image editors."""

import os
import shutil
import tempfile
import subprocess
import numpy as np

from . import params

from .image_io import load_image_as_array

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  def edit_with(self, command, export = "tiff", depth = 16, editor = "<Editor>", interactive = True, cwd = None):
    """Edit the image with an external tool.

    The image is saved on disk; the editor command is then run on this file, which is finally
    reloaded in eQuimageLab and returned.

    The user/editor must simply overwrite the edited file when leaving.

    Args:
      command (str): The command to be run (e.g., "gimp -n $").
        Any "$" is replaced by the name of the image file to be opened by the editor.
      export (str, optional): The format used to export the image. Can be:

        - "png": PNG file with depth = 8 or 16 bits integer per channel.
        - "tiff" (default): TIFF file with depth = 8, 16 or 32 bits integer per channel.
        - "fits": FITS file width depth = 32 bits float per channel.

      depth (int, optional): The color depth (bits per channel) used to export the image (see above; default 16).
      editor (str, optional): The name of the editor (for pretty-print purposes; default "<Editor>").
      interactive (bool, optional): If True (default), the editor is interactive (awaits commands from the user);
        if False, the editor processes the image autonomously and does not require inputs from the user.
      cwd (str, optional): If not None (default), change working directory to cwd before running the editor.

    Returns:
      Image: The edited image.
    """
    if export not in ["png", "tiff", "fits"]: raise ValueError(f"Error, unknown export format '{export}'.")
    if depth not in [8, 16, 32]: raise ValueError("Error, depth must be 8, 16 or 32 bpc.")
    # Edit image.
    with tempfile.TemporaryDirectory(ignore_cleanup_errors = True) as tmpdir:
      # Set tmp file name.
      filename = "eQuimageLab."+export
      filepath = os.path.join(tmpdir, filename)
      # Process command.
      splitcmd = []
      filefound = False
      for item in command.strip().split(" "):
        if item == "$":
          filefound = True
          splitcmd.append(filepath)
        else:
          splitcmd.append(item)
      if not filefound: raise ValueError("Error, no place holder for the image file ($) found in the command.")
      # Save image.
      print(f"Writing file {filepath} with depth = {depth} bpc...")
      self.save(filepath, depth = depth, compress = 0, verbose = False) # Don't compress image to ensure compatibility with the editor.
      ctime = os.path.getmtime(filepath)
      # Run editor.
      print(f"Running {editor}...")
      if interactive: print(f"Overwrite file {filepath} when leaving {editor}.")
      subprocess.run(splitcmd, cwd = cwd)
      # Load and return edited image.
      mtime = os.path.getmtime(filepath)
      if mtime == ctime:
        print(f"The image has not been modified by {editor}; Returning the original...")
        return self.copy()
      print(f"Reading file {filepath}...")
      image, meta = load_image_as_array(filepath, verbose = False)
      return self.newImage(image)

  def edit_with_gimp(self, export = "tiff", depth = 16):
    """Edit the image with Gimp.

    The image is saved on disk; Gimp is then run on this file, which is finally
    reloaded in eQuimageLab and returned.

    The user must simply overwrite the edited file when leaving Gimp.

    The command "gimp" must be in the PATH.

    Args:
      export (str, optional): The format used to export the image. Can be:

        - "png": PNG file with depth = 8 or 16 bits integer per channel.
        - "tiff" (default): TIFF file with depth = 8, 16 or 32 bits integer per channel.
        - "fits": FITS file width depth = 32 bits float per channel.

      depth (int, optional): The color depth (bits per channel) used to export the image (see above; default 16).

    Returns:
      Image: The edited image.
    """
    return self.edit_with("gimp -n $", export = export, depth = depth, editor = "Gimp", interactive = True)

  def edit_with_siril(self):
    """Edit the image with Siril.

    The image is saved as a FITS file (32 bits float per channel) on disk; Siril is then run on
    this file, which is finally reloaded in eQuimageLab and returned.

    The user must simply overwrite the edited file when leaving Siril.

    The command "siril" must be in the PATH.

    Returns:
      Image: The edited image.
    """
    return self.edit_with("siril $", export = "fits", depth = 32, editor = "Siril", interactive = True)

  def starnet(self, midtone = .5, starmask = False):
    """Remove the stars from the image with StarNet++.

    See: https://www.starnetastro.com/

    The image is saved as a TIFF file (16 bits integer per channel) on disk; the stars on this
    TIFF file are remove with StarNet++, and the starless image is finally reloaded in eQuimageLab
    and returned.

    The command "starnet++" must be in the PATH.

    Args:
      midtone (float, optional): If different from 0.5 (default), apply a midtone stretch to the
        image before running StarNet++, then apply the inverse stretch to the output starless.
        This can help StarNet++ find stars on low contrast, linear RGB images.
        See Image.midtone_stretch; midtone is expected in ]0, 1[.
      starmask (bool, optional): If True, return both the starless image and the star mask.
        If False (default), only return the starless image [the star mask being the difference
        between the original image (self) and the starless].

    Returns:
      Image: The starless image if starmask is False, and a tuple (starless image, star mask)
      if starmask is True.
    """
    # We need to cwd to the starnet++ directory to process the image.
    cmdpath = shutil.which("starnet++")
    if cmdpath is None: raise FileNotFoundError("Error, starnet++ executable not found in the PATH.")
    path, cmd = os.path.split(cmdpath)
    # Stretch the image if needed.
    if midtone != .5:
      image = self.midtone_stretch(midtone)
    else:
      image = self
    # Run starnet++.
    starless = image.edit_with("starnet++ $ $", export = "tiff", depth = 16, editor = "StarNet++", interactive = False, cwd = path)
    # "Unstretch" the starless if needed.
    if midtone != .5:
      starless = starless.midtone_stretch(midtone, inverse = True)
    # Return starless/star masks as appropriate.
    return (starless, self-starless) if starmask else starless



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image filters."""

import numpy as np
import scipy.ndimage as ndimg

from . import params

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  def sharpen(self, mode = "reflect", channels = ""):
    """Apply a sharpening (Laplacian) convolution filter to selected channels of the image.

    Args:
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd -> dcba|abcd|dcba).
        - "mirror": the image is reflected about the center of the last pixel (abcd -> dcb|abcd|cba).
        - "nearest": the image is padded with the value of the last pixel (abcd -> aaaa|abcd|dddd).
        - "zero": the image is padded with zeros (abcd -> 0000|abcd|0000).

    Returns:
      Image: The sharpened image.
    """
    # Translate modes.
    if mode == "zero": mode = "constant"
    # Set-up Laplacian kernel.
    kernel = np.array([[-1., -1., -1.], [-1., 9., -1.], [-1., -1., -1.]], dtype = params.IMGTYPE)
    # Convolve selected channels with the kernel.
    return self.apply_channels(lambda channel: ndimg.convolve(channel, kernel, mode = mode, cval = 0.), channels, multi = False)

  def remove_hot_pixels(self, ratio, mode = "reflect", channels = ""):
    """Remove hot pixels in selected channels of the image.

    All pixels of a selected channel greater than ratio times the eight nearest-neighbors average
    are replaced by this average.

    Args:
      ratio (float): The threshold for hot pixels detection.
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).

      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd -> dcba|abcd|dcba).
        - "mirror": the image is reflected about the center of the last pixel (abcd -> dcb|abcd|cba).
        - "nearest": the image is padded with the value of the last pixel (abcd -> aaaa|abcd|dddd).
        - "zero": the image is padded with zeros (abcd -> 0000|abcd|0000).

    Returns:
      Image: The processed image.
    """

    def remove_hot_pixels_channel(channel):
      """Remove hot pixels from the input channel data."""
      avg = ndimg.convolve(channel, kernel, mode = mode, cval = 0.)/nnn
      return np.where(channel > ratio*avg, avg, channel)

    if ratio <= 0.: raise ValueError("Error, ratio must be > 0.")
    # Translate modes.
    if mode == "zero": mode = "constant"
    # Set-up the (unnormalized) kernel for nearest-neighbors average.
    kernel = np.array([[1., 1., 1.], [1., 0., 1.], [1., 1., 1.]], dtype = params.IMGTYPE)
    # Normalize with respect to the actual number of nearest neighbors.
    nnn = ndimg.convolve(np.ones(*self.get_size()), kernel, mode = mode, cval = 0.)
    return self.apply_channels(remove_hot_pixels_channel, channels, multi = False)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Color management."""

import numpy as np
import scipy.interpolate as spint

from . import params
from . import helpers
from . import image_colorspaces as colorspaces

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  ##################
  # Image queries. #
  ##################

  def is_grayscale_RGB(self):
    """Return True if a RGB image is actually a grayscale (same RGB channels), False otherwise.

    Returns:
      bool: True if the RGB image is a grayscale (same RGB channels), False otherwise.
    """
    self.check_color_model("RGB")
    image = self.image
    return np.all(abs(image[1]-image[0]) < params.IMGTOL) and np.all(abs(image[2]-image[0]) < params.IMGTOL)

  ############################
  # Gray scales & negatives. #
  ############################

  def negative(self):
    """Return the negative of a RGB or grayscale image.

    Returns:
      Image: The negative of the image.
    """
    self.check_color_model("RGB", "gray")
    return 1.-self

  def grayscale(self, channel = "Y", RGB = False):
    """Convert the selected channel of a RGB image into a grayscale image.

    Args:
      channel: The converted channel ("V" for the HSV value, "L" for the luma, "Y" for the luminance).
        Namely, the output grayscale image has the same value/luma/luminance as the original RGB image.
      RGB (bool, optional): If True, return the grayscale as a RGB image (with identical R/G/B
        channels). If False (default), return the grayscale as a single channel image.

    Returns:
      Image: The grayscale image.
    """
    self.check_color_model("RGB")
    if channel == "V":
      grayscale = self.value()
    elif channel == "L":
      grayscale = self.luma()
    elif channel == "Y":
      grayscale = self.luminance()
      if self.colorspace == "sRGB":
        grayscale = colorspaces.lRGB_to_sRGB(grayscale) # Preserve luminance in the sRGB color space.
    else:
      raise ValueError(f"Error, invalid channel '{channel}' (must be 'V', 'L' or 'Y').")
    if RGB:
      return self.newImage(np.repeat(grayscale[np.newaxis, :, :], 3, axis = 0))
    else:
      return self.newImage(grayscale, colormodel = "gray")

  ##########################
  # Color transformations. #
  ##########################

  def color_balance(self, red = 1., green = 1., blue = 1.):
    """Adjust the color balance of a RGB image.

    Scales the red/green/blue channels by the input multipliers.

    Args:
      red (float, optional): The multiplier for the red channel (default 1).
      green (float, optional): The multiplier for the green channel (default 1).
      blue (float, optional): The multiplier for the blue channel (default 1).

    Returns:
      Image: The processed image.
    """
    self.check_color_model("RGB")
    if red < 0.: raise ValueError("Error, red must be >= 0.")
    if green < 0.: raise ValueError("Error, green must be >= 0.")
    if blue < 0.: raise ValueError("Error, blue must be >= 0.")
    output = self.copy()
    if red   != 1.: output.image[0] *= red
    if green != 1.: output.image[1] *= green
    if blue  != 1.: output.image[2] *= blue
    return output

  def color_saturation(self, A = 0., model = "midsat", interpolation = "cubic", trans = True, **kwargs):
    """Adjust color saturation.

    The image is converted to HSV (if needed) and the color saturation S is adjusted according to the 'model' kwarg:

      - "deltasat": Shift the saturation S ← S+delta.
      - "midsat": Apply a midtone stretch function S ← f(S) = (m-1)S/((2m-1)S-m) with midtone m = (1-delta)/2.
        This function increases monotonously from f(0) = 0 to f(m) = 1/2 and f(1) = 1.

    The image is converted back to the original color model ("RGB" or "HSV") after this operation.
    delta is expected in the [-1, 1] range, with delta = 0 leaving the image unchanged. delta > 0 saturates the colors,
    while delta < 0 turn the image into a a gray scale.
    delta is first set for all hues (with the 'A' kwarg), then can be updated for the red ('R'), yellow ('Y'), green ('G'),
    cyan ('C'), blue ('B') and magenta ('M') hues, if these kwargs are provided.
    delta is interpolated for arbitrary HSV hues using nearest neighbor, linear or cubic spline interpolation, according
    to the 'interpolation' kwarg.

    Args:
      A (float, optional): The delta for all hues (default 0).
      R (float, optional): The red delta (default A).
      Y (float, optional): The yellow delta (default A).
      G (float, optional): The green delta (default A).
      C (float, optional): The cyan delta (default A).
      B (float, optional): The blue delta (default A).
      M (float, optional): The magenta delta (default A).
      model (str, optional): The saturation model ["deltasat" or "midsat" (default)]
      interpolation (str, optional): The hue interpolation model:

        - "nearest": Nearest neighbor interpolation.
        - "linear": Linear spline interpolation.
        - "cubic": Cubic spline interpolation (default).

      trans (boolean, optional): If True, embeds the transformation delta(hue) in the output
        image as output.trans.

    Returns:
      Image: The processed image.
    """

    def interpolate(hue, psat, interpolation):
      """Interpolate the saturation parameter psat[RYGCBM] for arbitrary hues."""
      if np.all(psat == psat[0]):
        return np.full_like(hue, psat[0]) # Short-cut if the saturation parameter is the same for RYGCBM.
      hsat = np.linspace(0., 1., 7)
      psat = np.append(psat, psat[0]) # Enforce periodic boundary conditions.
      if interpolation == "nearest":
        fsat = spint.interp1d(hsat, psat, kind = "nearest")
      elif interpolation == "linear" or interpolation == "cubic":
        k = 3 if interpolation == "cubic" else 1
        tck = spint.splrep(hsat, psat, k = k, per = True) # Enforce periodic boundary conditions.
        def fsat(x): return np.clip(spint.splev(x, tck), -1., 1.)
      else:
        raise ValueError(f"Error, unknown interpolation method '{interpolation}'.")
      return fsat(hue)

    self.check_color_model("RGB", "HSV")
    psat = np.empty(6)
    psat[0] = kwargs.pop("R", A)
    psat[1] = kwargs.pop("Y", A)
    psat[2] = kwargs.pop("G", A)
    psat[3] = kwargs.pop("C", A)
    psat[4] = kwargs.pop("B", A)
    psat[5] = kwargs.pop("M", A)
    if kwargs: print("Discarding extra keyword arguments in Image.color_saturation...")
    hsv = self.HSV()
    hue = hsv.image[0]
    sat = hsv.image[1]
    delta = interpolate(hue, psat, interpolation)
    if model == "deltasat":
      sat += delta
    elif model == "midsat":
      midsat = np.clip(.5*(1.-delta), .005, .995)
      sat = (midsat-1.)*sat/((2.*midsat-1.)*sat-midsat)
    else:
      raise ValueError(f"Error, unknown saturation model '{model}.")
    hsv.image[1] = np.clip(sat, 0., 1.)
    output = hsv if self.colormodel == "HSV" else hsv.RGB()
    if trans:
      t = helpers.Container()
      t.type = "hue"
      t.input = self
      t.xm = np.linspace(0., 1., 7)
      t.ym = np.append(psat, psat[0])
      t.cm = ["red", "yellow", "green", "cyan", "blue", "magenta", "red"]
      t.x = np.linspace(0., 1., params.ntranslo)
      t.y = interpolate(t.x, psat, interpolation)
      t.ylabel = "\u0394"
      output.trans = t
    return output

  ##########################
  # Color noise reduction. #
  ##########################

  def SCNR(self, hue = "green", protection = "avgneutral", amount = 1., lightness = True):
    """Selective color noise reduction of a given hue of a RGB image.

    The input hue is reduced according to the 'protection' kwarg. For the green hue for example,

      - G ← min(G, C) with C = (R+B)/2 for average neutral protection (protection = "avgneutral").
      - G ← min(G, C) with C = max(R, B) for maximum neutral protection (protection = "maxneutral").
      - G ← G*[(1-A)+C*A] with C = (R+B)/2 for additive mask protection (protection = "addmask").
      - G ← G*[(1-A)+C*A] with C = max(R, B) for maximum mask protection (protection = "maxmask").

    The parameter A in [0, 1] controls the strength of the mask protection.

    Args:
      hue (str, optional): The hue to be reduced ["red" alias "R", "yellow" alias "Y", "green" alias "G" (default),
        "cyan" alias "C", "blue" alias "B", or "magenta" alias "M"].
      protection (str, optional): The protection mode ["avgneutral" (default), "maxneutral", "addmask" or "maxmask"].
      amount (float, optional): The parameter A for mask protection (protection = "addmask" or "maxmask", default 1).
      lightness (bool, optional): If True (default), rescale the output pixels to preserve the CIE lightness L* of
        the original image.

    Returns:
      Image: The processed image.
    """
    self.check_color_model("RGB")
    if hue == "red" or hue == "R":
      icc, ic1, ic2, negative = 0, 1, 2, False
    elif hue == "yellow" or hue == "Y":
      icc, ic1, ic2, negative = 2, 0, 1, True
    elif hue == "green" or hue == "G":
      icc, ic1, ic2, negative = 1, 0, 2, False
    elif hue == "cyan" or hue == "C":
      icc, ic1, ic2, negative = 0, 1, 2, True
    elif hue == "blue" or hue == "B":
      icc, ic1, ic2, negative = 2, 0, 1, False
    elif hue == "magenta" or hue == "M":
      icc, ic1, ic2, negative = 1, 0, 2, True
    else:
      raise ValueError(f"Error, unknown hue '{hue}'.")
    image = np.clip(self.image, 0., 1.) # Clip before reducing color noise.
    if negative: image = 1.-image
    if protection == "avgneutral":
      c = (image[ic1]+image[ic2])/2.
      image[icc] = np.minimum(image[icc], c)
    elif protection == "maxneutral":
      c = np.maximum(image[ic1], image[ic2])
      image[icc] = np.minimum(image[icc], c)
    elif protection == "addmask":
      c = np.minimum(1., image[ic1]+image[ic2])
      image[icc] *= (1.-amount)+c*amount
    elif protection == "maxmask":
      c = np.maximum(image[ic1], image[ic2])
      image[icc] *= (1.-amount)+c*amount
    else:
      raise ValueError(f"Error, unknown protection mode '{protection}'.")
    if negative: image = 1.-image
    output = self.newImage(image)
    if lightness:
      output.check_color_space("lRGB", "sRGB")
      output = output.lRGB()
      output = output.scale_pixels(output.luminance(), self.luminance())
      if self.colorspace == "sRGB": output = output.sRGB()
      difflight = output.lightness()-self.lightness()
      print(f"Maximum lightness difference = {abs(difflight).max()}.")
    return output



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image I/O management."""

import os
import numpy as np

from . import params

from PIL import Image as PILImage
if params.IMAGEIO:
  import imageio.v3 as iio
else:
  import skimage.io as skio
import astropy.io.fits as pyfits

def load_image_as_array(filename, verbose = True):
  """Load a RGB or grayscale image from a file.

  Args:
    filename (str): The file name.
    verbose (bool, optional): If True (default), print information about the image.

  Returns:
    The image as numpy.ndarray and the file meta-data (including exif if available) as a dictionary.
  """
  if verbose: print(f"Loading file {filename}...")
  try:
    header = PILImage.open(filename)
  except:
    header = None
    fmt = None
    if verbose: print("Failed to identify image file format; Attempting to load anyway...")
  else:
    fmt = header.format
    if verbose: print(f"Format = {fmt}.")
  if fmt == "PNG": # Load with the FreeImage plugin to enable 16 bits color depth.
    image = iio.imread(filename, plugin = "PNG-FI") if params.IMAGEIO else skio.imread(filename)
  elif fmt == "TIFF":
    image = iio.imread(filename, plugin = "TIFF") if params.IMAGEIO else skio.imread(filename, plugin = "tifffile")
  elif fmt == "FITS":
    hdus = pyfits.open(filename)
    image = hdus[0].data
    if image.ndim == 3:                 # Pyfits returns (channels, height, width)
      image = np.moveaxis(image, 0, -1) # instead of (height, width, channels),
    image = np.flip(image, axis = 0)    # and an upside down image.
  else:
    image = iio.imread(filename) if params.IMAGEIO else skio.imread(filename)
  if image.ndim == 2: # Assume single channel images are monochrome.
    nc = 1
    image = np.expand_dims(image, axis = -1)
  elif image.ndim == 3:
    nc = image.shape[2]
  else:
    raise ValueError(f"Error, invalid image shape = {image.shape}.")
  if verbose: print(f"Image size = {image.shape[1]}x{image.shape[0]} pixels.")
  if verbose: print(f"Number of channels = {nc}.")
  if nc not in [1, 3, 4]: raise ValueError(f"Error, images with {nc} channels are not supported.")
  dtype = str(image.dtype)
  if verbose: print(f"Data type = {dtype}.")
  if dtype == "uint8":
    bpc = 8
    image = params.IMGTYPE(image/255)
  elif dtype == "uint16":
    bpc = 16
    image = params.IMGTYPE(image/65535)
  elif dtype == "uint32":
    bpc = 32
    image = params.IMGTYPE(image/4294967295)
  elif dtype in ["float32", ">f4", "<f4"]: # Assumed normalized in [0, 1] !
    bpc = 32
    image = params.IMGTYPE(image)
  elif dtype in ["float64", ">f8", "<f8"]: # Assumed normalized in [0, 1] !
    bpc = 64
    image = params.IMGTYPE(image)
  else:
    raise TypeError(f"Error, image data type {dtype} is not supported.")
  if verbose: print(f"Bit depth per channel = {bpc}.")
  if verbose: print(f"Bit depth per pixel = {nc*bpc}.")
  image = np.moveaxis(image, -1, 0) # Move last (channel) axis to leading position.
  for ic in range(nc):
    if verbose: print(f"Channel #{ic}: minimum = {image[ic].min():.5f}, maximum = {image[ic].max():.5f}.")
  if nc == 4: image = image[0:3]*image[3] # Assume fourth channel is transparency.
  try:
    exif = header.getexif()
  except:
    exif = None
  else:
    if verbose: print("Succesfully read EXIF data...")
  meta = {"exif": exif, "colordepth": bpc}
  return np.ascontiguousarray(image), meta

def load_image(filename, colorspace = "sRGB", verbose = True):
  """Load a RGB or grayscale image from a file.

  Args:
    filename (str): The file name.
    colorspace (str, optional): The colorspace of the image [either "sRGB" (default) or "lRGB" for linear RGB images].
    verbose (bool, optional): If True (default), print information about the image.

  Returns:
    The image as an Image object and the file meta-data (including exif if available) as a dictionary.
  """
  from .image import Image
  image, meta = load_image_as_array(filename, verbose = verbose)
  return Image(image, colorspace = colorspace, colormodel = "RGB"), meta

def save_image(image, filename, depth = 8, compress = 5, verbose = True):
  """Save a RGB or grayscale image as a file.

  Note: The color space is *not* embedded in the file at present.

  Args:
    image (Image): The image.
    filename (str): The file name. The file format is chosen according to the extension:
      - .png: PNG file with depth = 8 or 16 bit integers per channel.
      - .tif, .tiff: TIFF file with depth = 8, 16 bit integers per channel, or 32 bit floats per channel.
      - .fit, .fits, .fts: FITS file with 32 bit floats per channel (irrespective of depth).
    depth (int, optional): The color depth of the file in bits per channel (default 8).
    compress (int, optional): The compression level for TIFF files (Default 5; 0 = no compression;
      actual compression if >0 depends whether ImageIO or scikit-image is used as I/O library -
      see `equimage.params.IMAGEIO`).
    verbose (bool, optional): If True (default), print information about the file.
  """
  image.check_color_model("RGB", "gray")
  is_gray = (image.colormodel == "gray")
  if is_gray:
    if verbose: print(f"Saving grayscale image as file {filename}...")
  else:
    if verbose: print(f"Saving RGB image as file {filename}...")
  root, ext = os.path.splitext(filename)
  if ext in [".png", ".tif", ".tiff"]:
    if depth == 8:
      image = image.int8()
    elif depth == 16:
      image = image.int16()
    elif depth == 32:
      image = image.int32()
    else:
      raise ValueError("Error, color depth must be 8 or 16, or 32 bits per channel.")
    if verbose: print(f"Color depth = {depth} bits integer per channel.")
    if is_gray: image = image[:, :, 0]
    if ext == ".png":
      if params.IMAGEIO:
        if depth > 16: raise ValueError("Error, color depth of png files must be 8 or 16 bits per channel.")
        iio.imwrite(filename, image, plugin = "PNG-FI")
      else:
        if depth > 8: raise ValueError("Error, color depth of png files must be 8 bits per channel.")
        skio.imsave(filename, image, check_contrast = False)
    elif ext == ".tif" or ext == ".tiff":
      if params.IMAGEIO:
        iio.imwrite(filename, image, plugin = "TIFF", metadata = {"compress": compress})
      else:
        skio.imsave(filename, image, plugin = "tifffile", check_contrast = False, compression = "zlib" if compress > 0 else None)
  elif ext in [".fit", ".fits", ".fts"]:
    if verbose: print(f"Color depth = {np.finfo(image.dtype).bits} bits float per channel.")
    image = image.flip_height().get_image() # Flip image upside down.
    if is_gray: image = image[0]
    hdu = pyfits.PrimaryHDU(image)
    hdu.writeto(filename, overwrite = True)
  else:
    raise ValueError("Error, file extension must be .png or .tif/.tiff., or .fit/.fits/.fts.")

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  def save(self, filename, depth = 8, compress = 5, verbose = True):
    """Save image as a file.

    Note: The color model must be "RGB" or "gray", but the color space is *not* embedded
    in the file at present.

    Args:
      filename (str): The file name. The file format is chosen according to the extension:
        - .png: PNG file with depth = 8 or 16 bits integer per channel.
        - .tif, .tiff: TIFF file with depth = 8, 16 bits integer per channel, or 32 bits float per channel.
        - .fit, .fits, .fts: FITS file with 32 bits float per channel (irrespective of depth).
      depth (int, optional): The color depth of the file in bits per channel (default 8).
      compress (int, optional): The compression level for TIFF files (Default 5; 0 = no compression;
        actual compression if >0 depends whether ImageIO or scikit-image is used as I/O library -
        see `equimage.params.IMAGEIO`).
      verbose (bool, optional): If True (default), print information about the file.
    """
    save_image(self, filename, depth = depth, compress = compress, verbose = verbose)
